import logging
from pprint import pformat
from flask import jsonify
from pymacaron_core.exceptions import ValidationError, PyMacaronCoreException
from pymacaron_core.swagger.apipool import ApiPool


log = logging.getLogger(__name__)


class PyMacaronException(PyMacaronCoreException):
    code = 'UNKNOWN_EXCEPTION'
    status = 500
    error_id = None
    user_message = None
    error_caught = None

    def tell_user(self, msg):
        self.user_message = msg
        return self

    def caught(self, error):
        self.error_caught = error
        return self

    def http_reply(self):
        """Return a Flask reply object describing this error"""
        data = {
            'status': self.status,
            'error': self.code.upper(),
            'error_description': str(self)
        }

        if self.error_caught:
            data['error_caught'] = pformat(self.error_caught)

        if self.error_id:
            data['error_id'] = self.error_id

        if self.user_message:
            data['user_message'] = self.user_message

        r = jsonify(data)
        r.status_code = self.status

        if str(self.status) != "200":
            log.warn("ERROR: caught error %s %s [%s]" % (self.status, self.code, str(self)))

        return r

    def to_model(self):
        """Return a bravado-core Error instance"""
        e = ApiPool().current_server_api.model.Error(
            status=self.status,
            error=self.code.upper(),
            error_description=str(self),
        )
        if self.error_id:
            e.error_id = self.error_id
        if self.user_message:
            e.user_message = self.user_message
        if self.error_caught:
            e.error_caught = pformat(self.error_caught)
        return e

#
# Interface to allow creating further Exception classes
#

code_to_class = {}

def add_error(name=None, code=None, status=None):
    """Create a new Exception class"""
    if not name or not status or not code:
        raise Exception("Can't create Exception class %s: you must set both name, status and code" % name)
    myexception = type(name, (PyMacaronException, ), {"code": code, "status": status})
    globals()[name] = myexception
    if code in code_to_class:
        raise Exception("ERROR! Exception %s is already defined." % code)
    code_to_class[code] = myexception
    return myexception


add_error('UnhandledServerError', 'UNHANDLED_SERVER_ERROR', 500)
add_error('InternalServerError', 'SERVER_ERROR', 500)
add_error('AuthMissingHeaderError', 'AUTHORIZATION_HEADER_MISSING', 401)
add_error('AuthTokenExpiredError', 'TOKEN_EXPIRED', 401)
add_error('AuthInvalidTokenError', 'TOKEN_INVALID', 401)
add_error('ValidationError', 'INVALID_PARAMETER', 400)

#
# Manipulate various error objects
#

def responsify(error):
    """Take an Error model and return it as a Flask response"""
    assert str(type(error).__name__) == 'Error'
    if error.error in code_to_class:
        e = code_to_class[error.error](error.error_description)
        if error.error_id:
            e.error_id = error.error_id
        if error.user_message:
            e.user_message = error.user_message
        return e.http_reply()
    elif isinstance(error, PyMacaronException):
        return error.http_reply()
    else:
        return PyMacaronException("Caught un-mapped error: %s" % error).http_reply()


def is_error(o):
    """True if o is an instance of a swagger Error model or a flask Response of
    an error model"""
    if hasattr(o, 'error') and hasattr(o, 'error_description') and hasattr(o, 'status'):
        return True
    return False


def format_error(e):
    """Take an exception caught within pymacaron_core and turn it into a
    bravado-core Error instance
    """

    if isinstance(e, PyMacaronException):
        return e.to_model()

    if isinstance(e, PyMacaronCoreException) and e.__class__.__name__ == 'ValidationError':
        return ValidationError(str(e)).to_model()

    # Turn this exception into a PyMacaron Error model
    return UnhandledServerError(str(e)).to_model()


def raise_error(e):
    """Take a bravado-core Error model and raise it as an exception"""
    code = e.error
    if code in code_to_class:
        raise code_to_class[code](e.error_description)
    else:
        raise InternalServerError(e.error_description)



import os
import logging
import json
import imp


utils = imp.load_source('utils', os.path.join(os.path.dirname(__file__), 'utils.py'))


log = logging.getLogger(__name__)


tmpdir = '/tmp/test-pym-microservice'
reportpath = os.path.join(tmpdir, "error_report.json")
try:
    os.stat(tmpdir)
except Exception:
    os.mkdir(tmpdir)


class Tests(utils.PyMacaronTests):


    def assertNoErrorReport(self):
        self.assertFalse(os.path.isfile(reportpath))

    def load_report(self):
        with open(reportpath) as f:
            s = f.read()
            log.info("GOT\n%s\n" % s)
            j = json.loads(s)
            log.debug("Report is %s" % json.dumps(j, indent=4))
            title = j['title']
            body = j['body']
            log.info("Loaded error report [%s]" % title)
            return title, body


    def assertBaseReportOk(self, path=None, user_id=None):
        title, body = self.load_report()
        self.assertTrue(title)
        # Urg. This is doomed to break for everyone else than me on my computer...
        self.assertTrue('/home/erwan/pnt/pymacaron/test/testserver.py' in body['stack'][0])
        self.assertTrue('call_id' in body)
        self.assertEqual(body['call_path'], 'crash')
        self.assertEqual(body['is_ec2_instance'], False)

        if user_id:
            self.assertEqual(body['user']['is_auth'], 1)
            self.assertEqual(body['user']['id'], user_id)
        else:
            self.assertEqual(body['user']['is_auth'], 0)
            self.assertEqual(body['user']['id'], '')
        self.assertTrue('python-requests' in body['user']['user_agent'])
        self.assertEqual(body['user']['ip'], '127.0.0.1')

        self.assertEqual(body['endpoint']['method'], 'GET')
        self.assertEqual(body['endpoint']['base_url'], 'http://127.0.0.1:8765/%s' % path)
        self.assertEqual(body['endpoint']['url'], 'http://127.0.0.1:8765/%s' % path)
        self.assertEqual(body['endpoint']['path'], '/%s' % path)
        self.assertTrue(body['endpoint']['id'].endswith(' GET /%s' % path), "%s ends with GET /%s" % (body['endpoint']['id'], path))

        self.assertEqual(body['server']['port'], '8765')
        self.assertTrue(body['server']['api_name'] in ['ping', 'crash'])
        self.assertEqual(body['server']['fqdn'], '127.0.0.1')
        self.assertEqual(body['server']['api_version'], '0.0.1')

        return title, body


    def assertServerErrorReportOk(self, path=None, fatal=True, user_id=None):
        title, body = self.assertBaseReportOk(user_id=user_id, path=path)
        self.assertTrue('error_id' in body)

        self.assertEqual(body['is_fatal_error'], fatal)

        self.assertTrue(body['time']['end'] >= body['time']['start'])
        self.assertTrue(body['time']['microsecs'] >= 1000)
        self.assertTrue(body['time']['microsecs'] <= 10000000)

        self.assertEqual(body['call_path'], 'crash')

        return title, body


    def setUp(self):
        super().setUp()

        if 'NO_ERROR_REPORTING' in os.environ:
            del os.environ['NO_ERROR_REPORTING']
        os.environ['DO_REPORT_ERROR'] = '1'

        if os.path.isfile(reportpath):
            os.remove(reportpath)

        self.kill_server()
        self.start_server()
        self.port = 8765

    #
    # And the tests!
    #

    def test_internal_exception(self):
        self.assertGetReturnError(
            'crash/internalexception',
            500,
            'UNHANDLED_SERVER_ERROR'
        )
        title, body = self.assertServerErrorReportOk(
            path='crash/internalexception',
        )
        self.assertEqual(title, 'FATAL ERROR %s 500 UNHANDLED_SERVER_ERROR: pymacaron.api.do_crash_internal_exception(): Raising an internal exception' % body['server']['api_name'])

        self.assertEqual(body['response']['user_message'], '')
        self.assertEqual(body['response']['type'], 'Response')
        self.assertEqual(body['response']['status'], '500')
        self.assertEqual(body['response']['is_error'], 1)
        self.assertEqual(body['response']['error_code'], 'UNHANDLED_SERVER_ERROR')
        self.assertEqual(body['response']['error_description'], 'Raising an internal exception')

        self.assertEqual(body['trace'][0], 'Traceback (most recent call last):\n')

        self.assertEqual(body['request']['params'], '[]')


    def test_pymacaron_exception(self):
        self.assertGetReturnError(
            'crash/pymacaronexception',
            401,
            'NON_FATAL_CUSTOM_ERROR'
        )
        self.assertNoErrorReport()


    def test_report_error(self):
        self.assertGetReturnOk(
            'crash/reporterror'
        )
        title, body = self.assertBaseReportOk(
            path='crash/reporterror',
        )
        self.assertEqual(title, 'NON-FATAL ERROR %s do_crash_report_error(): called crash/reporterror to test error reporting' % body['server']['api_name'])

        self.assertTrue('time' not in body)
        self.assertTrue('error_id' not in body)
        self.assertTrue('response' not in body)
        self.assertTrue('trace' not in body)
        self.assertTrue('request' not in body)
        self.assertEqual(body['is_fatal_error'], False)
        self.assertEqual(body['title'], 'called crash/reporterror to test error reporting')


    def test_report_fatal_error_response(self):
        self.assertGetReturnError(
            'crash/returnfatalerrorresponse',
            543,
            'FATAL_CUSTOM_ERROR'
        )
        title, body = self.assertServerErrorReportOk(
            path='crash/returnfatalerrorresponse',
            fatal=True,
        )
        self.assertEqual(title, 'FATAL ERROR %s 543 FATAL_CUSTOM_ERROR: pymacaron.api.do_crash_return_fatal_error_response(): endpoint returns an Error response' % body['server']['api_name'])


    def test_report_non_fatal_error_response(self):
        self.assertGetReturnError(
            'crash/returnnonfatalerrorresponse',
            401,
            'NON_FATAL_CUSTOM_ERROR'
        )
        self.assertNoErrorReport()


    def test_report_error_model(self):
        j = self.assertGetReturnError(
            'crash/returnerrormodel',
            543,
            'ANOTHER_CUSTOM_ERROR'
        )
        self.assertEqual(j['error_description'], 'Testing error model')
        self.assertEqual(j['status'], 543)
        self.assertEqual(j['error'], 'ANOTHER_CUSTOM_ERROR')
        title, body = self.assertServerErrorReportOk(
            path='crash/returnerrormodel',
            fatal=True,
        )
        self.assertEqual(title, 'FATAL ERROR %s 543 ANOTHER_CUSTOM_ERROR: pymacaron.api.do_crash_return_error_model(): Testing error model' % body['server']['api_name'])


    def test_report_error_instance(self):
        self.assertGetReturnError(
            'crash/returnerrorinstance',
            543,
            'FATAL_CUSTOM_ERROR'
        )
        title, body = self.assertServerErrorReportOk(
            path='crash/returnerrorinstance',
            fatal=True,
        )
        self.assertEqual(title, 'FATAL ERROR %s 543 FATAL_CUSTOM_ERROR: pymacaron.api.do_crash_return_error_instance(): endpoint returns an Error instance' % body['server']['api_name'])


    def test_report_slow_call(self):
        self.assertGetReturnOk(
            'crash/slowcall'
        )
        title, body = self.assertServerErrorReportOk(
            path='crash/slowcall',
            fatal=False,
        )
        self.assertEqual(title, 'NON-FATAL ERROR %s 200 : pymacaron.api.do_crash_slow_call() calltime exceeded 1000 millisec!' % body['server']['api_name'])

        self.assertEqual(body['response']['user_message'], '')
        self.assertEqual(body['response']['type'], 'Response')
        self.assertEqual(body['response']['status'], '200')
        self.assertEqual(body['response']['is_error'], 0)
        self.assertEqual(body['response']['error_code'], '')
        self.assertEqual(body['response']['error_description'], '')

        self.assertTrue('trace' not in body)

        self.assertEqual(body['request']['params'], '[]')

        self.assertEqual(body['title'], 'pymacaron.api.do_crash_slow_call() calltime exceeded 1000 millisec!')
